/*
 *  plugin.vala - Simple bracket completion plugin for gedit written
 *  in Vala
 *  
 *  Copyright (C) 2008 - Andrea Del Signore <sejerpz @ tin.it>
 *  
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *   
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330,
 *  Boston, MA 02111-1307, USA.
 */

using GLib;
using Gedit;
using Gdk;
using Gtk;

namespace Vala.GeditPlugins
{
	public class BracketCompletion : Gedit.Plugin
	{
		private string tab_chars = "";

		public override void activate (Gedit.Window window)
		{
			if (Gedit.prefs_manager_get_insert_spaces ())
				tab_chars = string.nfill (Gedit.prefs_manager_get_tabs_size (), ' ');
			else
				tab_chars = "\t";

			Signal.connect_after (window, "tab-added", (GLib.Callback) on_tab_added, this);
			Signal.connect_after (window, "tab-removed", (GLib.Callback) on_tab_removed, this);

			foreach (View view in window.get_views ()) {
				var doc = (Gedit.Document) (view.get_buffer ());
				if (doc.language != null && doc.language.id == "vala")
					connect_view (view);
			}

			foreach (Document doc in window.get_documents ()) {
				connect_document (doc);
			}
		}

		private void connect_view (Gedit.View view)
		{
			Signal.connect (view, "key-press-event", (GLib.Callback) on_view_key_press, this);
		}

		private void connect_document (Gedit.Document doc)
		{
			Signal.connect (doc, "notify::language", (GLib.Callback) on_notify_language, this);
		}

		private void disconnect_view (Gedit.View view)
		{
			SignalHandler.disconnect_by_func (view, (void*) on_view_key_press, this);
		}

		private void disconnect_document (Gedit.Document doc)
		{
			SignalHandler.disconnect_by_func (doc, (void*) on_notify_language, this);
		}

		public override void deactivate (Gedit.Window window)
		{
			foreach (View view in window.get_views ()) {
				disconnect_view (view);
			}
			foreach (Document doc in window.get_documents ()) {
				disconnect_document (doc);
			}

			SignalHandler.disconnect_by_func (window, (void*) on_tab_added, this);
			SignalHandler.disconnect_by_func (window, (void*) on_tab_removed, this);
		}


		public override bool is_configurable ()
		{
			return false;
		}

		private static void on_notify_language (Gedit.Document sender, ParamSpec pspec, BracketCompletion instance)
		{
			//search the view
			var app = App.get_default ();
			foreach (View view in app.get_views ()) {
				if (view.get_buffer () == sender) {
					if (sender.language  == null || sender.language.id != "vala") {
						instance.disconnect_view (view);
					} else {
						instance.connect_view (view);
					}
					break;
				}
			}
		}

		private static void on_tab_added (Gedit.Window sender, Gedit.Tab tab, BracketCompletion instance)
		{

			var doc = tab.get_document ();

			if (doc.language != null && doc.language.id == "vala") {
				var view = tab.get_view ();
				instance.connect_view (view);
			}
			instance.connect_document (doc);
		}

		private static void on_tab_removed (Gedit.Window sender, Gedit.Tab tab, BracketCompletion instance)
		{
			var view = tab.get_view ();
			var doc = tab.get_document ();

			instance.disconnect_view (view);
			instance.disconnect_document (doc);
		}

		private bool enclose_selection_with_delimiter (TextBuffer src, string delimiter)
		{
			TextIter sel_start;
			TextIter sel_end;
					
			src.get_selection_bounds (out sel_start, out sel_end);
			var text = src.get_text (sel_start, sel_end, true);
			if (text.has_prefix (delimiter) == false && text.has_suffix (delimiter) == false) {
				weak TextMark mark = (TextMark) src.get_insert ();
				TextIter pos;

				text = "%s%s%s".printf (delimiter, text, delimiter);
				src.begin_user_action ();
				src.delete_selection (true, true);
				src.get_iter_at_mark (out pos, mark);
				src.insert (pos, text, (int) text.len ());
				src.end_user_action ();
				return true;
			}

			return false;
		}

		private void insert_chars (TextBuffer src, string chars)
		{
			weak TextMark mark = (TextMark) src.get_insert ();
			TextIter pos;

			src.get_iter_at_mark (out pos, mark);
			src.begin_user_action ();
			src.insert (pos, chars, (int) chars.len ());
			src.end_user_action ();
		}

		private void move_backwards (TextBuffer src, int count)
		{
			weak TextMark mark = (TextMark) src.get_insert ();
			TextIter pos;

			src.get_iter_at_mark (out pos, mark);
			pos.backward_chars (count);
			src.place_cursor (pos);
		}

		private weak string current_indentation_text (TextBuffer src)
		{
			weak TextMark mark = (TextMark) src.get_insert ();
			TextIter end;
			TextIter start;
			weak string text = "";
			int col;

			src.get_iter_at_mark (out end, mark);
			col = end.get_line_offset ();
			end.set_line_offset (0);

			start = end;
			unichar ch = end.get_char ();

			if (ch.isspace ()) {
				while (end.forward_char ()) {
					if (end.starts_word ()) {
						break;
					} else if (end.get_line_offset () >= col) {
						break;
					} else {
						ch = end.get_char ();
						if (ch.isspace () == false) {
							break;
						}
					}
				}
			}

			if (!start.equal(end)) {
				text = start.get_text (end);
			}
			return text;
		}

		private static bool on_view_key_press (Gedit.View sender, Gdk.EventKey evt, BracketCompletion instance)
		{
			bool result = false;
	
			if ((evt.state & ( ModifierType.MOD1_MASK)) == 0) {
				var src = sender.get_buffer ();
				weak TextMark mark = (TextMark) src.get_insert ();
				TextIter pos;
				string text = evt.str;
				weak string indent;
				string buffer;

				src.get_iter_at_mark (out pos, mark);
				if (text == "(") {
					instance.insert_chars (src, ")");
					instance.move_backwards (src, 1);
				} else if (text == "{") {
				        indent = instance.current_indentation_text (src);
					buffer = "{\n%s%s\n%s}".printf(indent, instance.tab_chars, indent);
				        instance.insert_chars (src, buffer);
					instance.move_backwards (src, 2 + (int) indent.length);
					result = true;
				} else if (text == "\r" && (evt.state & ModifierType.SHIFT_MASK) != 0) {
					//move to end line
					indent = instance.current_indentation_text (src);
					pos.forward_to_line_end ();
					src.place_cursor (pos);
					if ((evt.state & ModifierType.CONTROL_MASK) == 0) {
						instance.insert_chars (src, ";\n%s".printf(indent));
						//place cursor to end
						src.get_iter_at_mark (out pos, mark);
						src.place_cursor (pos);
					} else {
						
						buffer = "\n%s{\n%s%s\n%s}".printf(indent, indent, instance.tab_chars, indent);
						instance.insert_chars (src, buffer);
						instance.move_backwards (src, 2 + (int) indent.length);
					}
					result = true;
				} else if (text == "\"" ) {
					if (src.has_selection) {
						if (instance.enclose_selection_with_delimiter (src, "\"")) {
							src.get_iter_at_mark (out pos, mark);
							src.place_cursor (pos);
							result = true;
						}
					} else {
						instance.insert_chars (src, "\"");
						instance.move_backwards (src, 1);
					}
				}

			}
			return result;
		}
	}
}

[ModuleInit]
public GLib.Type register_gedit_plugin (TypeModule module) 
{
	return typeof (Vala.GeditPlugins.BracketCompletion);
}